import httpx
from uuid import UUID
from typing import Optional, List
from pydantic import BaseModel
from contextlib import contextmanager
from enum import Enum


class OriginType(str, Enum):
    HUMAN = "human"
    AGENT = "agent"
    SCHEDULER = "scheduler"


class ResultType(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    PENDING = "pending"


class ActorType(str, Enum):
    AGENT = "agent"
    HUMAN = "human"


class ActionOption(BaseModel):
    action: str
    target: Optional[dict] = None
    score: float
    reason: Optional[str] = None


class IntentHandle:
    """Context manager for tracking an intent through its lifecycle."""

    def __init__(self, client: "AELClient", intent_id: UUID, session_id: str):
        self.client = client
        self.intent_id = intent_id
        self.session_id = session_id
        self.context_id: Optional[UUID] = None
        self.decision_id: Optional[UUID] = None

    def snapshot_context(
        self,
        inputs: dict,
        external_versions: Optional[dict] = None
    ) -> UUID:
        """Capture the context the agent sees."""
        response = self.client._post("/contexts", {
            "intent_id": str(self.intent_id),
            "inputs": inputs,
            "external_versions": external_versions or {}
        })
        self.context_id = UUID(response["id"])
        return self.context_id

    def decide(
        self,
        options: List[ActionOption],
        chosen_action: str,
        confidence: float,
        model_version: str = "unknown",
        rules_evaluated: Optional[List[str]] = None,
        reasoning: Optional[str] = None
    ) -> UUID:
        """Record a decision."""
        if not self.context_id:
            raise ValueError("Must call snapshot_context before decide")

        response = self.client._post("/decisions", {
            "intent_id": str(self.intent_id),
            "context_id": str(self.context_id),
            "options": [opt.model_dump() for opt in options],
            "chosen_action": chosen_action,
            "confidence": confidence,
            "model_version": model_version,
            "rules_evaluated": rules_evaluated or [],
            "reasoning": reasoning
        })
        self.decision_id = UUID(response["id"])
        return self.decision_id

    def execute(
        self,
        action: str,
        target: dict,
        result: ResultType = ResultType.SUCCESS,
        side_effects: Optional[List[str]] = None,
        revert_action: Optional[dict] = None,
        actor: ActorType = ActorType.AGENT,
        override_reason: Optional[str] = None
    ) -> UUID:
        """Record an execution."""
        if not self.decision_id:
            raise ValueError("Must call decide before execute")

        response = self.client._post("/executions", {
            "decision_id": str(self.decision_id),
            "action": action,
            "target": target,
            "result": result.value,
            "side_effects": side_effects or [],
            "revert_action": revert_action,
            "actor": actor.value,
            "override_reason": override_reason
        })
        return UUID(response["id"])


class AELClient:
    """Client for interacting with the Agent Execution Ledger API."""

    def __init__(
        self,
        endpoint: str = "http://localhost:8000",
        api_key: Optional[str] = None,
        timeout: float = 30.0
    ):
        self.endpoint = endpoint.rstrip("/")
        self.api_key = api_key
        self.timeout = timeout
        self._client = httpx.Client(timeout=timeout)

    def _headers(self) -> dict:
        headers = {"Content-Type": "application/json"}
        if self.api_key:
            headers["X-API-Key"] = self.api_key
        return headers

    def _post(self, path: str, data: dict) -> dict:
        url = f"{self.endpoint}/api/v1{path}"
        response = self._client.post(url, json=data, headers=self._headers())
        response.raise_for_status()
        return response.json()

    def _get(self, path: str) -> dict:
        url = f"{self.endpoint}/api/v1{path}"
        response = self._client.get(url, headers=self._headers())
        response.raise_for_status()
        return response.json()

    @contextmanager
    def intent(
        self,
        goal: str,
        agent_id: str = "default",
        session_id: Optional[str] = None,
        origin: OriginType = OriginType.AGENT,
        constraints: Optional[List[str]] = None
    ):
        """
        Context manager for tracking an agent intent.

        Usage:
            with ael.intent("Handle refund request") as intent:
                intent.snapshot_context({"ticket_id": 123})
                intent.decide(...)
                intent.execute(...)
        """
        import uuid as uuid_module

        if session_id is None:
            session_id = str(uuid_module.uuid4())

        response = self._post("/intents", {
            "goal": goal,
            "origin": origin.value,
            "constraints": constraints or [],
            "agent_id": agent_id,
            "session_id": session_id
        })

        intent_id = UUID(response["id"])
        handle = IntentHandle(self, intent_id, session_id)

        try:
            yield handle
        except Exception:
            # Could log failed intents here
            raise

    def track(self, goal: str, agent_id: str = "default"):
        """Decorator for tracking function executions as agent actions."""
        from .decorators import track as track_decorator
        return track_decorator(self, goal, agent_id)

    def get_execution(self, execution_id: UUID) -> dict:
        """Get execution details."""
        return self._get(f"/executions/{execution_id}")

    def get_session_timeline(self, session_id: str) -> dict:
        """Get full timeline for a session."""
        return self._get(f"/sessions/{session_id}/timeline")

    def replay(self, execution_id: UUID, with_context_id: Optional[UUID] = None) -> dict:
        """Replay a decision."""
        data = {}
        if with_context_id:
            data["with_context_id"] = str(with_context_id)
        return self._post(f"/replay/{execution_id}", data)

    def revert(self, execution_id: UUID, reason: str, force: bool = False) -> dict:
        """Revert an execution."""
        return self._post(f"/revert/{execution_id}", {
            "reason": reason,
            "force": force
        })

    def close(self):
        """Close the HTTP client."""
        self._client.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Google ADK (Agent Development Kit) integration for AEL.

This module provides utilities to track Google ADK agent actions in AEL.
"""

from typing import Any, Optional
from uuid import uuid4
from contextlib import contextmanager

from ..client import AELClient, ActionOption, ResultType, ActorType, OriginType


class AELTracker:
    """
    Tracker for Google ADK agents that records actions in AEL.

    Usage with Google ADK:
        from google import genai
        from google.adk import Agent
        from ael import AELClient
        from ael.integrations.google_adk import AELTracker

        # Initialize
        ael = AELClient(endpoint="http://localhost:8000", api_key="ael_xxx")
        tracker = AELTracker(ael, agent_id="support-agent")

        # Create your ADK agent
        agent = Agent(name="support", model="gemini-2.0-flash", ...)

        # Track agent interactions
        with tracker.track("Handle refund request") as t:
            t.context({"customer_id": "123", "request": "refund"})

            # Run your agent
            response = agent.run(user_input)

            # Record what happened
            t.decision(
                chosen="approve_refund",
                confidence=0.92,
                reasoning=response.text,
                options=[
                    {"action": "approve_refund", "score": 0.92},
                    {"action": "deny_refund", "score": 0.08},
                ]
            )
            t.execute(action="approve_refund", result="success")
    """

    def __init__(
        self,
        client: AELClient,
        agent_id: str = "google-adk-agent",
        model_version: str = "gemini-2.0-flash"
    ):
        self.client = client
        self.agent_id = agent_id
        self.model_version = model_version
        self._current_session: Optional[str] = None

    def start_session(self, session_id: Optional[str] = None) -> str:
        """Start a new tracking session."""
        self._current_session = session_id or str(uuid4())
        return self._current_session

    @contextmanager
    def track(self, goal: str, session_id: Optional[str] = None):
        """
        Context manager for tracking an agent action.

        Usage:
            with tracker.track("Handle customer request") as t:
                t.context({"input": "..."})
                # ... run agent ...
                t.decision(chosen="action", confidence=0.9, ...)
                t.execute(action="action", result="success")
        """
        session = session_id or self._current_session or str(uuid4())
        tracker = _ActionTracker(
            client=self.client,
            goal=goal,
            agent_id=self.agent_id,
            session_id=session,
            model_version=self.model_version
        )
        try:
            yield tracker
        finally:
            tracker._finalize()

    def track_action(
        self,
        goal: str,
        inputs: dict,
        chosen_action: str,
        confidence: float,
        reasoning: str,
        options: Optional[list[dict]] = None,
        result: str = "success",
        side_effects: Optional[list[str]] = None,
    ) -> dict:
        """
        One-shot tracking of a complete agent action.

        Args:
            goal: What the agent was trying to accomplish
            inputs: The context/inputs the agent received
            chosen_action: The action the agent chose
            confidence: Confidence score (0-1)
            reasoning: Why the agent made this decision
            options: List of options considered [{"action": str, "score": float, "reason": str}]
            result: "success", "failure", or "pending"
            side_effects: List of side effects

        Returns:
            Dict with execution details
        """
        session_id = self._current_session or str(uuid4())

        with self.client.intent(
            goal=goal,
            agent_id=self.agent_id,
            session_id=session_id,
            origin=OriginType.AGENT
        ) as intent:
            intent.snapshot_context(inputs, {"model": self.model_version})

            # Build options
            action_options = []
            if options:
                for opt in options:
                    action_options.append(ActionOption(
                        action=opt.get("action", "unknown"),
                        score=opt.get("score", 0.5),
                        reason=opt.get("reason")
                    ))
            else:
                action_options.append(ActionOption(
                    action=chosen_action,
                    score=confidence
                ))

            intent.decide(
                options=action_options,
                chosen_action=chosen_action,
                confidence=confidence,
                model_version=self.model_version,
                reasoning=reasoning
            )

            result_type = {
                "success": ResultType.SUCCESS,
                "failure": ResultType.FAILURE,
                "pending": ResultType.PENDING
            }.get(result, ResultType.SUCCESS)

            execution_id = intent.execute(
                action=chosen_action,
                target=inputs,
                result=result_type,
                side_effects=side_effects or [],
                actor=ActorType.AGENT
            )

            return {
                "execution_id": str(execution_id),
                "session_id": session_id
            }


class _ActionTracker:
    """Internal tracker for a single action within a track() context."""

    def __init__(self, client, goal, agent_id, session_id, model_version):
        self._client = client
        self._goal = goal
        self._agent_id = agent_id
        self._session_id = session_id
        self._model_version = model_version
        self._context_data = {}
        self._decision_data = None
        self._execution_data = None
        self._intent_handle = None

    def context(self, inputs: dict, external_versions: Optional[dict] = None):
        """Record the context the agent saw."""
        self._context_data = {
            "inputs": inputs,
            "external_versions": external_versions or {"model": self._model_version}
        }

    def decision(
        self,
        chosen: str,
        confidence: float,
        reasoning: str = "",
        options: Optional[list[dict]] = None,
        rules_evaluated: Optional[list[str]] = None
    ):
        """Record the decision made by the agent."""
        self._decision_data = {
            "chosen": chosen,
            "confidence": confidence,
            "reasoning": reasoning,
            "options": options or [{"action": chosen, "score": confidence}],
            "rules_evaluated": rules_evaluated or []
        }

    def execute(
        self,
        action: str,
        result: str = "success",
        target: Optional[dict] = None,
        side_effects: Optional[list[str]] = None,
        revert_action: Optional[dict] = None
    ):
        """Record the execution result."""
        self._execution_data = {
            "action": action,
            "result": result,
            "target": target or self._context_data.get("inputs", {}),
            "side_effects": side_effects or [],
            "revert_action": revert_action
        }

    def _finalize(self):
        """Finalize and send all tracking data to AEL."""
        if not self._decision_data:
            return  # Nothing to track

        with self._client.intent(
            goal=self._goal,
            agent_id=self._agent_id,
            session_id=self._session_id,
            origin=OriginType.AGENT
        ) as intent:
            # Context
            intent.snapshot_context(
                self._context_data.get("inputs", {}),
                self._context_data.get("external_versions", {})
            )

            # Decision
            options = [
                ActionOption(
                    action=opt.get("action", "unknown"),
                    score=opt.get("score", 0.5),
                    reason=opt.get("reason")
                )
                for opt in self._decision_data.get("options", [])
            ]

            intent.decide(
                options=options,
                chosen_action=self._decision_data["chosen"],
                confidence=self._decision_data["confidence"],
                model_version=self._model_version,
                reasoning=self._decision_data.get("reasoning", ""),
                rules_evaluated=self._decision_data.get("rules_evaluated", [])
            )

            # Execution
            if self._execution_data:
                result_map = {
                    "success": ResultType.SUCCESS,
                    "failure": ResultType.FAILURE,
                    "pending": ResultType.PENDING
                }
                intent.execute(
                    action=self._execution_data["action"],
                    target=self._execution_data["target"],
                    result=result_map.get(self._execution_data["result"], ResultType.SUCCESS),
                    side_effects=self._execution_data.get("side_effects", []),
                    revert_action=self._execution_data.get("revert_action"),
                    actor=ActorType.AGENT
                )


# Backward compatibility alias
AELCallback = AELTracker



from .client import (
    AELClient,
    ActionOption,
    OriginType,
    ResultType,
    ActorType,
)
from .decorators import track
from .context import ContextBuilder, snapshot_context

__all__ = [
    "AELClient",
    "ActionOption",
    "OriginType",
    "ResultType",
    "ActorType",
    "track",
    "ContextBuilder",
    "snapshot_context",
]
__version__ = "0.1.1"



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="ael-sdk",
    version="0.1.1",
    author="Vinay Badhan",
    author_email="vinay.badhan21.work@gmail.com",
    description="Python SDK for Agent Execution Ledger - Audit trail for AI agents",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vinayb21-work/Agent-Execution-Ledger",
    project_urls={
        "Bug Tracker": "https://github.com/vinayb21-work/Agent-Execution-Ledger/issues",
        "Documentation": "https://github.com/vinayb21-work/Agent-Execution-Ledger#readme",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Logging",
    ],
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "httpx>=0.26.0",
        "pydantic>=2.5.0",
    ],
    extras_require={
        "google-adk": ["google-genai>=0.3.0"],
    },
    keywords="ai agents audit logging llm observability",
)



import hashlib
import json
from datetime import datetime
from typing import Any, Optional


class ContextBuilder:
    """
    Helper class for building context snapshots.

    Usage:
        ctx = ContextBuilder()
        ctx.add("ticket", ticket_data)
        ctx.add("customer", customer_data)
        ctx.set_version("crm", "v2.1")
        ctx.set_version("policy", "2026-01-01")

        with ael.intent("Handle ticket") as intent:
            intent.snapshot_context(ctx.inputs, ctx.versions)
    """

    def __init__(self):
        self._inputs: dict[str, Any] = {}
        self._versions: dict[str, str] = {}
        self._timestamp = datetime.utcnow().isoformat()

    def add(self, key: str, value: Any) -> "ContextBuilder":
        """Add an input to the context."""
        self._inputs[key] = value
        return self

    def add_all(self, data: dict) -> "ContextBuilder":
        """Add multiple inputs from a dictionary."""
        self._inputs.update(data)
        return self

    def set_version(self, system: str, version: str) -> "ContextBuilder":
        """Set the version of an external system."""
        self._versions[system] = version
        return self

    @property
    def inputs(self) -> dict:
        """Get all inputs."""
        return self._inputs

    @property
    def versions(self) -> dict:
        """Get all versions."""
        return self._versions

    def compute_hash(self) -> str:
        """Compute a hash of the context for integrity verification."""
        data = {
            "inputs": self._inputs,
            "versions": self._versions,
            "timestamp": self._timestamp
        }
        serialized = json.dumps(data, sort_keys=True, default=str)
        return hashlib.sha256(serialized.encode()).hexdigest()


def snapshot_context(
    inputs: dict,
    external_versions: Optional[dict] = None
) -> tuple[dict, dict]:
    """
    Simple function to prepare context data for snapshot.

    Returns:
        Tuple of (inputs, external_versions) ready for snapshot_context call
    """
    return inputs, external_versions or {}


def serialize_for_context(obj: Any) -> Any:
    """
    Serialize an object for inclusion in context.
    Handles common types that need special treatment.
    """
    if hasattr(obj, "model_dump"):
        # Pydantic model
        return obj.model_dump()
    elif hasattr(obj, "__dict__"):
        # Regular object
        return {k: v for k, v in obj.__dict__.items() if not k.startswith("_")}
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, (list, tuple)):
        return [serialize_for_context(item) for item in obj]
    elif isinstance(obj, dict):
        return {k: serialize_for_context(v) for k, v in obj.items()}
    else:
        return obj



import functools
import inspect
from typing import Callable, Optional
from uuid import uuid4

from .client import AELClient, ActionOption, ResultType, ActorType


def track(
    client: AELClient,
    goal: str,
    agent_id: str = "default",
    model_version: str = "unknown"
):
    """
    Decorator to automatically track function executions in the AEL.

    Usage:
        @ael.track(goal="Process customer refund")
        def handle_refund(ticket_id: str, amount: float):
            # Your agent logic here
            return {"status": "approved", "amount": amount}

    The decorator will:
    1. Create an intent with the specified goal
    2. Snapshot the function arguments as context
    3. Record a decision based on the function execution
    4. Record the execution result
    """
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            session_id = str(uuid4())

            # Get function signature to capture all args
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            inputs = dict(bound_args.arguments)

            with client.intent(
                goal=goal,
                agent_id=agent_id,
                session_id=session_id
            ) as intent:
                # Snapshot context (function inputs)
                intent.snapshot_context(inputs)

                # Execute the function
                try:
                    result = func(*args, **kwargs)
                    success = True
                    error = None
                except Exception as e:
                    result = None
                    success = False
                    error = str(e)

                # Record decision (we treat the function as a single decision)
                action_name = func.__name__
                intent.decide(
                    options=[
                        ActionOption(
                            action=action_name,
                            target=inputs,
                            score=1.0 if success else 0.0,
                            reason="Function executed" if success else f"Error: {error}"
                        )
                    ],
                    chosen_action=action_name,
                    confidence=1.0 if success else 0.0,
                    model_version=model_version,
                    reasoning=f"Executed {action_name} with args: {list(inputs.keys())}"
                )

                # Record execution
                intent.execute(
                    action=action_name,
                    target={"args": inputs, "result": result if success else None},
                    result=ResultType.SUCCESS if success else ResultType.FAILURE,
                    side_effects=[f"called_{action_name}"],
                    actor=ActorType.AGENT
                )

                if not success:
                    raise Exception(error)

                return result

        return wrapper
    return decorator
